using System;
using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;


namespace LarrysArms
{
	public class HairColorChanger : MVRScript
	{
		#region Variables
		// Basics
		Atom _person;
		List<DAZHairGroup> _activeHairGroups;

		// Path for presets
		string _presetPath = Application.dataPath.Replace("/VaM_Data", "") + "/Custom/Hair/Female/";

		// Main Buttons
		UIDynamicButton _btnUpdateHairList;
		JSONStorableStringChooser _chooserHairStorable;
		UIDynamicButton _btnApplyToAll;
		UIDynamicButton _btnRestoreHairDefaults;

		// Toggle Buttons
		Color _defaultBtnColor;
		UIDynamicButton _btnHairColorToggle;
		bool _hairColorUiEnabled;
		UIDynamicButton _btnScalpColorToggle;
		bool _scalpColorUiEnabled;

		// Preset file browser
		JSONStorableUrl _presetFileUrl;
		UIDynamicButton _btnLoadPreset;

		// Hair color properties
		JSONStorableFloat _floatPrimarySpecGloss;
		JSONStorableFloat _floatSecondarySpecGloss;
		JSONStorableFloat _floatSecondarySpecShift;
		JSONStorableFloat _floatRandomColorPower;
		JSONStorableFloat _floatRandomColorOffset;
		JSONStorableFloat _floatDiffuseSoftness;
		JSONStorableFloat _floatFresnelPower;
		JSONStorableFloat _floatFresnelAttenuation;
		JSONStorableFloat _floatGiFactor;
		JSONStorableFloat _floatNormalRandomize;
		JSONStorableColor _colorHairRoot;
		JSONStorableColor _colorHairTip;
		JSONStorableFloat _floatRootTipRolloff;
		JSONStorableColor _colorHairSpec;

		// Hair color UI
		UIDynamicSlider _sliderPrimarySpecGloss;
		UIDynamicSlider _sliderSecondarySpecGloss;
		UIDynamicSlider _sliderSecondarySpecShift;
		UIDynamicSlider _sliderRandomColorPower;
		UIDynamicSlider _sliderRandomColorOffset;
		UIDynamicSlider _sliderDiffuseSoftness;
		UIDynamicSlider _sliderFresnelPower;
		UIDynamicSlider _sliderFresnelAttenuation;
		UIDynamicSlider _sliderGiFactor;
		UIDynamicSlider _sliderNormalRandomize;
		UIDynamic _hairColorSpacer;
		UIDynamicColorPicker _colorPickerHairRoot;
		UIDynamicColorPicker _colorPickerHairTip;
		UIDynamicSlider _sliderRootTipRolloff;
		UIDynamicColorPicker _colorPickerHairSpec;
		UIDynamicButton _btnMaterialSliderDefaults;
		UIDynamicButton _btnMaterialColorDefaults;

		// Scalp color properties
		JSONStorableFloat _floatSpecularTextureOffset;
		JSONStorableFloat _floatSpecularIntensity;
		JSONStorableFloat _floatGloss;
		JSONStorableFloat _floatSpecularFresnel;
		JSONStorableFloat _floatGlossTextureOffset;
		JSONStorableFloat _floatGiFilter;
		JSONStorableFloat _floatAlphaAdjust;
		JSONStorableFloat _floatDiffuseTextureOffset;
		JSONStorableColor _colorScalpDiffuse;
		JSONStorableColor _colorScalpSpecular;
		JSONStorableColor _colorScalpSubsurface;

		// Scalp color UI
		UIDynamicSlider _sliderSpecularTextureOffset;
		UIDynamicSlider _sliderSpecularIntensity;
		UIDynamicSlider _sliderGloss;
		UIDynamicSlider _sliderSpecularFresnel;
		UIDynamicSlider _sliderGlossTextureOffset;
		UIDynamicSlider _sliderGiFilter;
		UIDynamicSlider _sliderAlphaAdjust;
		UIDynamicSlider _sliderDiffuseTextureOffset;
		UIDynamicColorPicker _colorPickerScalpDiffuse;
		UIDynamicColorPicker _colorPickerScalpSpecular;
		UIDynamicColorPicker _colorPickerScalpSubsurface;
#endregion

		public override void Init()
		{
			try
			{
				JSONStorableAction TrueUpdate = new JSONStorableAction("True Update", () =>
				{
					UpdateActiveHairGroups();
				}); 
				_person = SuperController.singleton.GetAtomByUid("Person");
				if (_person.category != "People")
				{
					SuperController.LogError("This didn't work.");
					return;
				}

				RegisterAction(TrueUpdate);
				SetupHairProperties();
				SetupScalpProperties();

				UpdateActiveHairGroups();   // Update hair list
				CreateUI(); // Create UI
			}
			catch (Exception e)
			{
				SuperController.LogError("Exception caught: " + e);
			}
		}

		protected void UpdateActiveHairGroups()
		{
			_activeHairGroups = new List<DAZHairGroup>();

			foreach (DAZHairGroup hairGroup in _person.GetComponentsInChildren<DAZHairGroup>())
			{
				DAZHairGroupControl hairControl = hairGroup.GetComponentInChildren<DAZHairGroupControl>();
				HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();

				if (hairControl != null && simControl != null)
				{
					_activeHairGroups.Add(hairGroup);
				}
			}
		}

		private void SetupHairProperties()
		{
			// Hair colour material info properties
			_floatPrimarySpecGloss = new JSONStorableFloat("Primary Specular Gloss", 160.00f, HairMaterialFloatCallback, 0, 256.00f);
			_floatSecondarySpecGloss = new JSONStorableFloat("Secondary Specular Gloss", 64.00f, HairMaterialFloatCallback, 0, 256.0f);
			_floatSecondarySpecShift = new JSONStorableFloat("Secondary Specular Shift", 0.400f, HairMaterialFloatCallback, 0, 1.0f);
			_floatRandomColorPower = new JSONStorableFloat("Random Color Power", 2.000f, HairMaterialFloatCallback, 0, 10.0f);
			_floatRandomColorOffset = new JSONStorableFloat("Random Color Offset", 0.300f, HairMaterialFloatCallback, 0, 1.0f);
			_floatDiffuseSoftness = new JSONStorableFloat("Diffuse Softness", 0.100f, HairMaterialFloatCallback, 0, 1.0f);
			_floatFresnelPower = new JSONStorableFloat("Fresnel Power", 8.00f, HairMaterialFloatCallback, 0, 10.0f);
			_floatFresnelAttenuation = new JSONStorableFloat("Fresnel Attenuation", 0.200f, HairMaterialFloatCallback, 0, 1.0f);
			_floatGiFactor = new JSONStorableFloat("GI Factor", 0.500f, HairMaterialFloatCallback, 0, 1.0f);
			_floatNormalRandomize = new JSONStorableFloat("Normal Randomize", 0.00f, HairMaterialFloatCallback, 0, 1.0f);
			_floatRootTipRolloff = new JSONStorableFloat("Root Tip Rolloff", 3f, HairMaterialFloatCallback, 0f, 5f, true);
			RegisterFloat(_floatPrimarySpecGloss);
			RegisterFloat(_floatSecondarySpecGloss);
			RegisterFloat(_floatSecondarySpecShift);
			RegisterFloat(_floatRandomColorPower);
			RegisterFloat(_floatRandomColorOffset);
			RegisterFloat(_floatDiffuseSoftness);
			RegisterFloat(_floatFresnelPower);
			RegisterFloat(_floatFresnelAttenuation);
			RegisterFloat(_floatGiFactor);
			RegisterFloat(_floatNormalRandomize);
			RegisterFloat(_floatRootTipRolloff);

			// Default hair colours
			HSVColor rootColor; rootColor.H = 0.079f; rootColor.S = 0.91f; rootColor.V = 0.09f;
			HSVColor tipColor; tipColor.H = 0.079f; tipColor.S = 0.68f; tipColor.V = 0.57f;
			HSVColor specColor; specColor.H = 0f; specColor.S = 0f; specColor.V = 0.67f;

			// Colour properties
			_colorHairRoot = new JSONStorableColor("Root Color", rootColor, HairColorCallback);
			_colorHairTip = new JSONStorableColor("Tip Color", tipColor, HairColorCallback);
			_colorHairSpec = new JSONStorableColor("Spec Color", specColor, HairColorCallback);
			RegisterColor(_colorHairRoot);
			RegisterColor(_colorHairTip);
			RegisterColor(_colorHairSpec);
		}

		private void SetupScalpProperties()
		{
			// Scalp material info properties
			_floatSpecularTextureOffset = new JSONStorableFloat("Specular Texture Offset", 0f, ScalpMaterialFloatCallback, -1.0f, 1.0f);
			_floatSpecularIntensity = new JSONStorableFloat("Specular Intensity", 0f, ScalpMaterialFloatCallback, 0, 10.0f);
			_floatGloss = new JSONStorableFloat("Gloss", 3.200f, ScalpMaterialFloatCallback, 2f, 8f);
			_floatSpecularFresnel = new JSONStorableFloat("Specular Fresnel", 0.5f, ScalpMaterialFloatCallback, 0, 1.0f);
			_floatGlossTextureOffset = new JSONStorableFloat("Gloss Texture Offset", 0f, ScalpMaterialFloatCallback, -1f, 1f);
			_floatGiFilter = new JSONStorableFloat("Global Illumination Filter", 0.5f, ScalpMaterialFloatCallback, 0, 1.0f);
			_floatAlphaAdjust = new JSONStorableFloat("Alpha Adjust", 0f, ScalpMaterialFloatCallback, -1f, 1.0f);
			_floatDiffuseTextureOffset = new JSONStorableFloat("Diffuse Texture Offset", 0f, ScalpMaterialFloatCallback, -1f, 1f);
			RegisterFloat(_floatSpecularTextureOffset);
			RegisterFloat(_floatSpecularIntensity);
			RegisterFloat(_floatGloss);
			RegisterFloat(_floatSpecularFresnel);
			RegisterFloat(_floatGlossTextureOffset);
			RegisterFloat(_floatGiFilter);
			RegisterFloat(_floatAlphaAdjust);
			RegisterFloat(_floatDiffuseTextureOffset);

			// Default scalp colours
			HSVColor diffuseColor; diffuseColor.H = 0.0f; diffuseColor.S = 0.24f; diffuseColor.V = 0.13f;
			HSVColor specColor; specColor.H = 0.0f; specColor.S = 0.24f; specColor.V = 0.13f;
			HSVColor subsurfaceColor; subsurfaceColor.H = 0f; subsurfaceColor.S = 0f; subsurfaceColor.V = 0f;

			// Colour properties
			_colorScalpDiffuse = new JSONStorableColor("Diffuse Color", diffuseColor, ScalpColorCallback);
			_colorScalpSpecular = new JSONStorableColor("Specular Color", specColor, ScalpColorCallback);
			_colorScalpSubsurface = new JSONStorableColor("Subsurface Color", subsurfaceColor, ScalpColorCallback);
			RegisterColor(_colorScalpDiffuse);
			RegisterColor(_colorScalpSpecular);
			RegisterColor(_colorScalpSubsurface);
		}

		protected void CreateUI()
		{
			// Update Hair List button
			_btnUpdateHairList = CreateButton("Update Hair List");
			_btnUpdateHairList.button.onClick.AddListener(UpdateHairListCallback);

			// Hair chooser drop-down
			_chooserHairStorable = new JSONStorableStringChooser("hairStorableChooser", GetHairNames(), "None", "Copy from: ", HairStorableChooserCallback);
			UIDynamicPopup udp = CreatePopup(_chooserHairStorable);

			// Apply To All button
			_btnApplyToAll = CreateButton("Apply To All");
			_btnApplyToAll.button.onClick.AddListener(ApplyToAllButtonCallback);

			// Restore Hair Defaults button
			_btnRestoreHairDefaults = CreateButton("Restore Hair Defaults");
			_btnRestoreHairDefaults.button.onClick.AddListener(RestoreHairDefaultsButtonCallback);

			// Plugin info text field
			JSONStorableString stringVersion = new JSONStorableString("versionNum", "\nHair Color Changer 1.2 by LarrysArms.");
			UIDynamicTextField textFieldVersion = CreateTextField(stringVersion, true);
			textFieldVersion.height = 110;

			CreateSpacer(true).height = 20;

			// Preset file browser
			_btnLoadPreset = CreateButton("Load Preset ", true);
			_presetFileUrl = new JSONStorableUrl("Preset Browser", _presetPath, FileBrowserCallback, "vap");
			_presetFileUrl.suggestedPath = _presetPath;

			_presetFileUrl.RegisterFileBrowseButton(_btnLoadPreset.button);

			CreateSpacer(true).height = 50;

			// Colour options toggle buttons
			_btnHairColorToggle = CreateButton("Change Hair Color");
			_btnHairColorToggle.button.onClick.AddListener(HairColorUIButtonCallback);
			_btnScalpColorToggle = CreateButton("Change Scalp Color", true);
			_btnScalpColorToggle.button.onClick.AddListener(ScalpColorUIButtonCallback);

			_defaultBtnColor = _btnHairColorToggle.buttonColor;

			ToggleButtonStatus();
		}

		private void ToggleButtonStatus()
		{
			bool activeHairsAvailable = _activeHairGroups.Count > 0;

			_btnApplyToAll.button.interactable = activeHairsAvailable;
			_btnRestoreHairDefaults.button.interactable = activeHairsAvailable;
			_btnLoadPreset.button.interactable = activeHairsAvailable;
			_btnHairColorToggle.button.interactable = activeHairsAvailable;
			_btnScalpColorToggle.button.interactable = activeHairsAvailable;

			if (!activeHairsAvailable)
			{
				if (_hairColorUiEnabled)
					DisableHairColorUI();

				if (_scalpColorUiEnabled)
					DisableScalpColorUI();
			}
		}

		protected void FileBrowserCallback(string file)
		{
			var jsonInfo = SuperController.singleton.LoadJSON(file);

			JSONNode storables = jsonInfo["storables"];
			JSONNode hairProperties = null;
			JSONNode scalpProperties = null;

			foreach (var storable in storables.Childs)
			{
				if (storable["id"].Value.EndsWith("Sim"))
				{
					hairProperties = storable;
				}
				else if (storable["id"].Value.EndsWith("ScalpMaterial"))
				{
					scalpProperties = storable;
				}
			}

			if (hairProperties != null)
			{
				JSONNode rootColorNode = hairProperties["rootColor"];
				JSONNode tipColorNode = hairProperties["tipColor"];
				JSONNode specColorNode = hairProperties["specularColor"];

				HSVColor rootColor;
				rootColor.H = rootColorNode["h"].AsFloat; rootColor.S = rootColorNode["s"].AsFloat; rootColor.V = rootColorNode["v"].AsFloat;
				HSVColor tipColor;
				tipColor.H = tipColorNode["h"].AsFloat; tipColor.S = tipColorNode["s"].AsFloat; tipColor.V = tipColorNode["v"].AsFloat;
				HSVColor specColor;
				specColor.H = specColorNode["h"].AsFloat; specColor.S = specColorNode["s"].AsFloat; specColor.V = specColorNode["v"].AsFloat;

				_floatPrimarySpecGloss.valNoCallback = hairProperties["primarySpecularSharpness"].AsFloat;
				_floatSecondarySpecGloss.valNoCallback = hairProperties["secondarySpecularSharpness"].AsFloat;
				_floatSecondarySpecShift.valNoCallback = hairProperties["specularShift"].AsFloat;
				_floatRandomColorPower.valNoCallback = hairProperties["randomColorPower"].AsFloat;
				_floatRandomColorOffset.valNoCallback = hairProperties["randomColorOffset"].AsFloat;
				_floatDiffuseSoftness.valNoCallback = hairProperties["diffuseSoftness"].AsFloat;
				_floatFresnelPower.valNoCallback = hairProperties["fresnelPower"].AsFloat;
				_floatFresnelAttenuation.valNoCallback = hairProperties["fresnelAttenuation"].AsFloat;
				_floatGiFactor.valNoCallback = hairProperties["IBLFactor"].AsFloat;
				_floatNormalRandomize.valNoCallback = hairProperties["normalRandomize"].AsFloat;

				_colorHairRoot.valNoCallback = rootColor;
				_colorHairTip.valNoCallback = tipColor;
				_floatRootTipRolloff.valNoCallback = hairProperties["colorRolloff"].AsFloat;
				_colorHairSpec.valNoCallback = specColor;

				UpdateAllHairProperties();
			}

			if (scalpProperties != null)
			{
				JSONNode diffuseColorNode = scalpProperties["Diffuse Color"];
				JSONNode specularColorNode = scalpProperties["Specular Color"];
				JSONNode subsurfaceColorNode = scalpProperties["Subsurface Color"];

				HSVColor diffuseColor;
				diffuseColor.H = diffuseColorNode["h"].AsFloat; diffuseColor.S = diffuseColorNode["s"].AsFloat; diffuseColor.V = diffuseColorNode["v"].AsFloat;
				HSVColor specularColor;
				specularColor.H = specularColorNode["h"].AsFloat; specularColor.S = specularColorNode["s"].AsFloat; specularColor.V = specularColorNode["v"].AsFloat;
				HSVColor subsurfaceColor;
				subsurfaceColor.H = subsurfaceColorNode["h"].AsFloat; subsurfaceColor.S = subsurfaceColorNode["s"].AsFloat; subsurfaceColor.V = subsurfaceColorNode["v"].AsFloat;

				_floatSpecularTextureOffset.valNoCallback = scalpProperties["Specular Texture Offset"].AsFloat;
				_floatSpecularIntensity.valNoCallback = scalpProperties["Specular Intensity"].AsFloat;
				_floatGloss.valNoCallback = scalpProperties["Gloss"].AsFloat;
				_floatSpecularFresnel.valNoCallback = scalpProperties["Specular Fresnel"].AsFloat;
				_floatGlossTextureOffset.valNoCallback = scalpProperties["Gloss Texture Offset"].AsFloat;
				_floatGiFilter.valNoCallback = scalpProperties["Global Illumination Filter"].AsFloat;
				_floatAlphaAdjust.valNoCallback = scalpProperties["Alpha Adjust"].AsFloat;
				_floatDiffuseTextureOffset.valNoCallback = scalpProperties["Diffuse Texture Offset"].AsFloat;

				_colorScalpDiffuse.valNoCallback = diffuseColor;
				_colorScalpSpecular.valNoCallback = specularColor;
				_colorScalpSubsurface.valNoCallback = subsurfaceColor;

				UpdateAllScalpProperties();
			}
		}

		protected List<string> GetHairNames()
		{
			List<string> hairNames = new List<string>();

			foreach (DAZHairGroup hairGroup in _activeHairGroups)
			{
				hairNames.Add(hairGroup.displayName);
			}

			return hairNames;
		}

		public void UpdateHairListCallback()
		{
			UpdateActiveHairGroups();
			ToggleButtonStatus();
			_chooserHairStorable.SetVal("");
			_chooserHairStorable.choices = GetHairNames();
		}

		private void HairColorUIButtonCallback()
		{
			_hairColorUiEnabled = !_hairColorUiEnabled;

			if (_hairColorUiEnabled)
			{
				EnableHairColorUI();
				DisableScalpColorUI();
			}
			else
			{
				DisableHairColorUI();
			}
		}

		private void ScalpColorUIButtonCallback()
		{
			_scalpColorUiEnabled = !_scalpColorUiEnabled;

			if (_scalpColorUiEnabled)
			{
				EnableScalpColorUI();
				DisableHairColorUI();
			}
			else
			{
				DisableScalpColorUI();
			}
		}

		private void EnableHairColorUI()
		{
			_hairColorUiEnabled = true;
			_btnHairColorToggle.buttonColor = Color.green;

			// Material info left-hand side
			_sliderPrimarySpecGloss = CreateSlider(_floatPrimarySpecGloss);
			_sliderSecondarySpecGloss = CreateSlider(_floatSecondarySpecGloss);
			_sliderSecondarySpecShift = CreateSlider(_floatSecondarySpecShift);
			_sliderRandomColorPower = CreateSlider(_floatRandomColorPower);
			_sliderRandomColorOffset = CreateSlider(_floatRandomColorOffset);
			_sliderDiffuseSoftness = CreateSlider(_floatDiffuseSoftness);
			_sliderFresnelPower = CreateSlider(_floatFresnelPower);
			_sliderFresnelAttenuation = CreateSlider(_floatFresnelAttenuation);
			_sliderGiFactor = CreateSlider(_floatGiFactor);
			_sliderNormalRandomize = CreateSlider(_floatNormalRandomize);

			// Material info right-hand side
			_colorPickerHairRoot = CreateColorPicker(_colorHairRoot, true);
			_colorPickerHairTip = CreateColorPicker(_colorHairTip, true);
			_sliderRootTipRolloff = CreateSlider(_floatRootTipRolloff, true);
			_colorPickerHairSpec = CreateColorPicker(_colorHairSpec, true);

			_hairColorSpacer = CreateSpacer(true);
			_hairColorSpacer.height = 15;

			// Reset Material Slider defaults button
			_btnMaterialSliderDefaults = CreateButton("Reset Material Slider defaults");
			_btnMaterialSliderDefaults.button.onClick.AddListener(ResetMaterialSliderCallback);

			// Reset Color Picker defaults button
			_btnMaterialColorDefaults = CreateButton("Reset Color Picker defaults", true);
			_btnMaterialColorDefaults.button.onClick.AddListener(ResetColorPickerCallback);
		}

		private void DisableHairColorUI()
		{
			_hairColorUiEnabled = false;
			_btnHairColorToggle.buttonColor = _defaultBtnColor;

			// Material info left-hand side
			RemoveSlider(_sliderPrimarySpecGloss);
			RemoveSlider(_sliderSecondarySpecGloss);
			RemoveSlider(_sliderSecondarySpecShift);
			RemoveSlider(_sliderRandomColorPower);
			RemoveSlider(_sliderRandomColorOffset);
			RemoveSlider(_sliderDiffuseSoftness);
			RemoveSlider(_sliderFresnelPower);
			RemoveSlider(_sliderFresnelAttenuation);
			RemoveSlider(_sliderGiFactor);
			RemoveSlider(_sliderNormalRandomize);

			// Material info right-hand side
			RemoveColorPicker(_colorPickerHairRoot);
			RemoveColorPicker(_colorPickerHairTip);
			RemoveSlider(_sliderRootTipRolloff);
			RemoveColorPicker(_colorPickerHairSpec);

			RemoveSpacer(_hairColorSpacer);

			// Reset Material Slider defaults button
			_btnMaterialSliderDefaults.button.onClick.RemoveListener(ResetMaterialSliderCallback);
			RemoveButton(_btnMaterialSliderDefaults);

			// Reset Color Picker defaults button
			_btnMaterialColorDefaults.button.onClick.RemoveListener(ResetColorPickerCallback);
			RemoveButton(_btnMaterialColorDefaults);
		}

		private void EnableScalpColorUI()
		{
			_scalpColorUiEnabled = true;
			_btnScalpColorToggle.buttonColor = Color.green;

			_sliderSpecularTextureOffset = CreateSlider(_floatSpecularTextureOffset);
			_sliderSpecularIntensity = CreateSlider(_floatSpecularIntensity);
			_sliderGloss = CreateSlider(_floatGloss);
			_sliderSpecularFresnel = CreateSlider(_floatSpecularFresnel);
			_sliderGlossTextureOffset = CreateSlider(_floatGlossTextureOffset);
			_sliderGiFilter = CreateSlider(_floatGiFilter);
			_sliderAlphaAdjust = CreateSlider(_floatAlphaAdjust);
			_sliderDiffuseTextureOffset = CreateSlider(_floatDiffuseTextureOffset);

			// Material info right-hand side
			_colorPickerScalpDiffuse = CreateColorPicker(_colorScalpDiffuse, true);
			_colorPickerScalpSpecular = CreateColorPicker(_colorScalpSpecular, true);
			_colorPickerScalpSubsurface = CreateColorPicker(_colorScalpSubsurface, true);
		}

		private void DisableScalpColorUI()
		{
			_scalpColorUiEnabled = false;
			_btnScalpColorToggle.buttonColor = _defaultBtnColor;

			RemoveSlider(_sliderSpecularTextureOffset);
			RemoveSlider(_sliderSpecularIntensity);
			RemoveSlider(_sliderGloss);
			RemoveSlider(_sliderSpecularFresnel);
			RemoveSlider(_sliderGlossTextureOffset);
			RemoveSlider(_sliderGiFilter);
			RemoveSlider(_sliderAlphaAdjust);
			RemoveSlider(_sliderDiffuseTextureOffset);

			// Material info right-hand side
			RemoveColorPicker(_colorPickerScalpDiffuse);
			RemoveColorPicker(_colorPickerScalpSpecular);
			RemoveColorPicker(_colorPickerScalpSubsurface);
		}

		public void ApplyToAllButtonCallback()
		{
			UpdateAllHairProperties();
			UpdateAllScalpProperties();
		}

		public void RestoreHairDefaultsButtonCallback()
		{
			foreach (var hairGroup in _activeHairGroups)
			{
				HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();
				DAZSkinWrapMaterialOptions scalpControl = hairGroup.GetComponentInChildren<DAZSkinWrapMaterialOptions>();

				if (simControl != null && scalpControl != null)
				{
					_chooserHairStorable.SetVal("");
					simControl.RestoreAllFromDefaults();
					scalpControl.RestoreAllFromDefaults();

					GetCurrentHairMaterialProperties(simControl);
					GetCurrentScalpMaterialProperties(scalpControl);
				}
			}
		}

		protected void HairMaterialFloatCallback(float value)
		{
			UpdateAllHairProperties();
		}

		protected void ScalpMaterialFloatCallback(float value)
		{
			UpdateAllScalpProperties();
		}

		protected void HairColorCallback(JSONStorableColor jcolor)
		{
			UpdateAllHairProperties();
		}

		protected void ScalpColorCallback(JSONStorableColor jcolor)
		{
			UpdateAllScalpProperties();
		}

		private void UpdateAllHairProperties()
		{
			foreach (var hairGroup in _activeHairGroups)
			{
				HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();

				if (simControl != null)
				{
					SetCurrentHairMaterialProperties(simControl);
				}
			}
		}

		private void UpdateAllScalpProperties()
		{
			foreach (var hairGroup in _activeHairGroups)
			{
				DAZSkinWrapMaterialOptions scalpControl = hairGroup.GetComponentInChildren<DAZSkinWrapMaterialOptions>();

				if (scalpControl != null)
				{
					SetCurrentScalpMaterialProperties(scalpControl);
				}
			}
		}

		protected void ResetMaterialSliderCallback()
		{
			_floatPrimarySpecGloss.defaultVal = 160f;
			_floatSecondarySpecGloss.defaultVal = 64f;
			_floatSecondarySpecShift.defaultVal = 0.4f;
			_floatRandomColorPower.defaultVal = 2.0f;
			_floatRandomColorOffset.defaultVal = 0.3f;
			_floatDiffuseSoftness.defaultVal = 0.1f;
			_floatFresnelPower.defaultVal = 8.0f;
			_floatFresnelAttenuation.defaultVal = 0.2f;
			_floatGiFactor.defaultVal = 0.5f;
			_floatNormalRandomize.defaultVal = 0.0f;

			_floatRootTipRolloff.defaultVal = 3.0f;

			_floatPrimarySpecGloss.SetValToDefault();
			_floatSecondarySpecGloss.SetValToDefault();
			_floatSecondarySpecShift.SetValToDefault();
			_floatRandomColorPower.SetValToDefault();
			_floatRandomColorOffset.SetValToDefault();
			_floatDiffuseSoftness.SetValToDefault();
			_floatFresnelPower.SetValToDefault();
			_floatFresnelAttenuation.SetValToDefault();
			_floatGiFactor.SetValToDefault();
			_floatNormalRandomize.SetValToDefault();

			_floatRootTipRolloff.SetValToDefault();
		}

		protected void ResetColorPickerCallback()
		{
			HSVColor rootColor; rootColor.H = 0.079f; rootColor.S = 0.91f; rootColor.V = 0.09f;
			HSVColor tipColor; tipColor.H = 0.079f; tipColor.S = 0.68f; tipColor.V = 0.57f;
			HSVColor specColor; specColor.H = 0f; specColor.S = 0f; specColor.V = 0.67f;

			_colorHairRoot.defaultVal = rootColor;
			_colorHairTip.defaultVal = tipColor;
			_colorHairSpec.defaultVal = specColor;

			_colorHairRoot.SetValToDefault();
			_colorHairTip.SetValToDefault();
			_colorHairSpec.SetValToDefault();
		}

		protected void HairStorableChooserCallback(string hairStorableID)
		{
			foreach (var hairGroup in _activeHairGroups)
			{
				if (hairStorableID == hairGroup.displayName)
				{
					HairSimControl simControl = hairGroup.GetComponentInChildren<HairSimControl>();
					DAZSkinWrapMaterialOptions scalpControl = hairGroup.GetComponentInChildren<DAZSkinWrapMaterialOptions>();

					if (simControl != null)
					{
						GetCurrentHairMaterialProperties(simControl);

						// Set slider defaults to material defaults
						_floatPrimarySpecGloss.SetDefaultFromCurrent();
						_floatSecondarySpecGloss.SetDefaultFromCurrent();
						_floatSecondarySpecShift.SetDefaultFromCurrent();
						_floatRandomColorPower.SetDefaultFromCurrent();
						_floatRandomColorOffset.SetDefaultFromCurrent();
						_floatDiffuseSoftness.SetDefaultFromCurrent();
						_floatFresnelPower.SetDefaultFromCurrent();
						_floatFresnelAttenuation.SetDefaultFromCurrent();
						_floatGiFactor.SetDefaultFromCurrent();
						_floatNormalRandomize.SetDefaultFromCurrent();

						_colorHairRoot.SetDefaultFromCurrent();
						_colorHairTip.SetDefaultFromCurrent();
						_floatRootTipRolloff.SetDefaultFromCurrent();
						_colorHairSpec.SetDefaultFromCurrent();
					}

					if (scalpControl != null)
					{
						GetCurrentScalpMaterialProperties(scalpControl);

						_floatSpecularTextureOffset.SetDefaultFromCurrent();
						_floatSpecularIntensity.SetDefaultFromCurrent();
						_floatGloss.SetDefaultFromCurrent();
						_floatSpecularFresnel.SetDefaultFromCurrent();
						_floatGlossTextureOffset.SetDefaultFromCurrent();
						_floatGiFilter.SetDefaultFromCurrent();
						_floatAlphaAdjust.SetDefaultFromCurrent();
						_floatDiffuseTextureOffset.SetDefaultFromCurrent();

						_colorScalpDiffuse.SetDefaultFromCurrent();
						_colorScalpSpecular.SetDefaultFromCurrent();
						_colorScalpSubsurface.SetDefaultFromCurrent();
					}
				}
			}
		}

		protected void SetCurrentHairMaterialProperties(HairSimControl hairProperties)
		{
			hairProperties.SetFloatParamValue("primarySpecularSharpness", _floatPrimarySpecGloss.val);
			hairProperties.SetFloatParamValue("secondarySpecularSharpness", _floatSecondarySpecGloss.val);
			hairProperties.SetFloatParamValue("specularShift", _floatSecondarySpecShift.val);
			hairProperties.SetFloatParamValue("randomColorPower", _floatRandomColorPower.val);
			hairProperties.SetFloatParamValue("randomColorOffset", _floatRandomColorOffset.val);
			hairProperties.SetFloatParamValue("diffuseSoftness", _floatDiffuseSoftness.val);
			hairProperties.SetFloatParamValue("fresnelPower", _floatFresnelPower.val);
			hairProperties.SetFloatParamValue("fresnelAttenuation", _floatFresnelAttenuation.val);
			hairProperties.SetFloatParamValue("IBLFactor", _floatGiFactor.val);
			hairProperties.SetFloatParamValue("normalRandomize", _floatNormalRandomize.val);

			hairProperties.SetColorParamValue("rootColor", _colorHairRoot.val);
			hairProperties.SetColorParamValue("tipColor", _colorHairTip.val);
			hairProperties.SetFloatParamValue("colorRolloff", _floatRootTipRolloff.val);
			hairProperties.SetColorParamValue("specularColor", _colorHairSpec.val);
		}

		protected void GetCurrentHairMaterialProperties(HairSimControl hairProperties)
		{
			_floatPrimarySpecGloss.valNoCallback = hairProperties.GetFloatParamValue("primarySpecularSharpness");
			_floatSecondarySpecGloss.valNoCallback = hairProperties.GetFloatParamValue("secondarySpecularSharpness");
			_floatSecondarySpecShift.valNoCallback = hairProperties.GetFloatParamValue("specularShift");
			_floatRandomColorPower.valNoCallback = hairProperties.GetFloatParamValue("randomColorPower");
			_floatRandomColorOffset.valNoCallback = hairProperties.GetFloatParamValue("randomColorOffset");
			_floatDiffuseSoftness.valNoCallback = hairProperties.GetFloatParamValue("diffuseSoftness");
			_floatFresnelPower.valNoCallback = hairProperties.GetFloatParamValue("fresnelPower");
			_floatFresnelAttenuation.valNoCallback = hairProperties.GetFloatParamValue("fresnelAttenuation");
			_floatGiFactor.valNoCallback = hairProperties.GetFloatParamValue("IBLFactor");
			_floatNormalRandomize.valNoCallback = hairProperties.GetFloatParamValue("normalRandomize");

			_colorHairRoot.valNoCallback = hairProperties.GetColorParamValue("rootColor");
			_colorHairTip.valNoCallback = hairProperties.GetColorParamValue("tipColor");
			_floatRootTipRolloff.valNoCallback = hairProperties.GetFloatParamValue("colorRolloff");
			_colorHairSpec.valNoCallback = hairProperties.GetColorParamValue("specularColor");
		}

		protected void SetCurrentScalpMaterialProperties(DAZSkinWrapMaterialOptions scalpProperties)
		{
			scalpProperties.SetFloatParamValue("Specular Texture Offset", _floatSpecularTextureOffset.val);
			scalpProperties.SetFloatParamValue("Specular Intensity", _floatSpecularIntensity.val);
			scalpProperties.SetFloatParamValue("Gloss", _floatGloss.val);
			scalpProperties.SetFloatParamValue("Specular Fresnel", _floatSpecularFresnel.val);
			scalpProperties.SetFloatParamValue("Gloss Texture Offset", _floatGlossTextureOffset.val);
			scalpProperties.SetFloatParamValue("Global Illumination Filter", _floatGiFilter.val);
			scalpProperties.SetFloatParamValue("Alpha Adjust", _floatAlphaAdjust.val);
			scalpProperties.SetFloatParamValue("Diffuse Texture Offset", _floatDiffuseTextureOffset.val);
			
			scalpProperties.SetColorParamValue("Diffuse Color", _colorScalpDiffuse.val);
			scalpProperties.SetColorParamValue("Specular Color", _colorScalpSpecular.val);
			scalpProperties.SetColorParamValue("Subsurface Color", _colorScalpSubsurface.val);
		}

		protected void GetCurrentScalpMaterialProperties(DAZSkinWrapMaterialOptions scalpProperties)
		{
			_floatSpecularTextureOffset.valNoCallback = scalpProperties.GetFloatParamValue("Specular Texture Offset");
			_floatSpecularIntensity.valNoCallback = scalpProperties.GetFloatParamValue("Specular Intensity");
			_floatGloss.valNoCallback = scalpProperties.GetFloatParamValue("Gloss");
			_floatSpecularFresnel.valNoCallback = scalpProperties.GetFloatParamValue("Specular Fresnel");
			_floatGlossTextureOffset.valNoCallback = scalpProperties.GetFloatParamValue("Gloss Texture Offset");
			_floatGiFilter.valNoCallback = scalpProperties.GetFloatParamValue("Global Illumination Filter");
			_floatAlphaAdjust.valNoCallback = scalpProperties.GetFloatParamValue("Alpha Adjust");
			_floatDiffuseTextureOffset.valNoCallback = scalpProperties.GetFloatParamValue("Diffuse Texture Offset");

			_colorScalpDiffuse.valNoCallback = scalpProperties.GetColorParamValue("Diffuse Color");
			_colorScalpSpecular.valNoCallback = scalpProperties.GetColorParamValue("Specular Color");
			_colorScalpSubsurface.valNoCallback = scalpProperties.GetColorParamValue("Subsurface Color");
		}
	}
}